/* ========================================================================== */
/*   GTuner                                                                   */
/*   (c) 2015 SP                                                              */
/*                                                                            */
/*   Program for Guitar Tuner with ATtiny44A (24) and 8x2 LCD                 */
/*   Clock frequency: Xtal 20 MHz                                             */
/* ========================================================================== */
// Version: 1.0 - first released (AVR Studio 4.18)
#define F_CPU 20000000UL

#include <avr/io.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#define 	wdt_reset()   __asm__ __volatile__ ("wdr")

// Global variables (for quicker and smaller code)
long measval,sumval;	// measured value, sum total for 4 measval
unsigned int freqres;	// Global variable with frequency
unsigned char updown, tonerange, text[9]={"Tuner\0"}; // Arrow for tuning, index nearest tone in table

// Tone names (2 chars only for each tone, missig 's' will be completed)
uint8_t names[] = " CCi DDi E FFi GGi AAi H";
// Octave index
uint8_t octind[] = {2,0x20,0x20,3,4};

// Graphic table for user characters in CGRAM in the LCD
prog_uint8_t specchars[] = {
// Up arrow
0b00000100,
0b00001110,
0b00011111,
0b00001110,
0b00001110,
0b00001110,
0b00001110,
0b00000000,
// Down arrow
0b00001110,
0b00001110,
0b00001110,
0b00001110,
0b00011111,
0b00001110,
0b00000100,
0b00000000,
// Lower 1
0b00000000,
0b00000000,
0b00000010,
0b00000110,
0b00000010,
0b00000010,
0b00000010,
0b00000000,
// Upper 1
0b00000010,
0b00000110,
0b00000010,
0b00000010,
0b00000010,
0b00000000,
0b00000000,
0b00000000,
// Upper 2
0b00000110,
0b00001001,
0b00000010,
0b00000100,
0b00001111,
0b00000000,
0b00000000,
0b00000000,
};

// 1 octave (kontra) tone table (number of 2,5 MHz impulses in 32 periods)
prog_uint32_t tones[12] = {2446244,2308949,2179356,2057036,1941587,1832614,
1729756,1632673,1541037,1454545,1372907,1295853};

// ******** Bit manipulation definitions ***********
#define ONTUNER asm ("sbi 0x18,2" ::)
#define OFFTUNER asm ("cbi 0x18,2" ::)
// LCD 2x8 chars, Write only, 4 bits bus
#define DISP_E_0 asm ("cbi 0x1B,7" ::)  // LCD signals definition
#define DISP_E_1 asm ("sbi 0x1B,7" ::)
#define DISP_INST asm ("cbi 0x1B,6" ::)
#define DISP_DATA asm ("sbi 0x1B,6" ::)
#define DISP_DB7_0 asm ("cbi 0x1B,0" ::)
#define DISP_DB7_1 asm ("sbi 0x1B,0" ::)
#define DISP_DB6_0 asm ("cbi 0x1B,1" ::)
#define DISP_DB6_1 asm ("sbi 0x1B,1" ::)
#define DISP_DB5_0 asm ("cbi 0x1B,4" ::)
#define DISP_DB5_1 asm ("sbi 0x1B,4" ::)
#define DISP_DB4_0 asm ("cbi 0x1B,5" ::)
#define DISP_DB4_1 asm ("sbi 0x1B,5" ::)
// Delay for display instructions
#define Delay1u _delay_us(1)
#define Delay50u _delay_us(50)
#define Delay2ms _delay_ms(2)

// Display type - define MAXCHARS x MAXLINES
#define MAXCHARS 8 
#define MAXLINES 2 

// ************** Display ******************
// Write data to display - 4 bits bus
void WriteCom(uint8_t codedat)
{
	if(codedat & 0x80) DISP_DB7_1; else DISP_DB7_0;
	if(codedat & 0x40) DISP_DB6_1; else DISP_DB6_0;
	if(codedat & 0x20) DISP_DB5_1; else DISP_DB5_0;
	if(codedat & 0x10) DISP_DB4_1; else DISP_DB4_0;
	// Write upper nibble
	DISP_E_1;
	Delay1u;
	DISP_E_0;
	if(codedat & 0x08) DISP_DB7_1; else DISP_DB7_0;
	if(codedat & 0x04) DISP_DB6_1; else DISP_DB6_0;
	if(codedat & 0x02) DISP_DB5_1; else DISP_DB5_0;
	if(codedat & 0x01) DISP_DB4_1; else DISP_DB4_0;
	// Write lower nibble
	DISP_E_1;
	Delay1u;
	DISP_E_0;
	Delay50u;
}

// Write instruction code to display
void WriteIns(uint8_t codein)
{
	DISP_INST;	// Set display for instruction
	WriteCom(codein);
}

// Write data (char) to display
void WriteChar(uint8_t codein)
{
	DISP_DATA;	// Set display for data
	WriteCom(codein);
}

void InitDisplay()
{
// Display is 8bit after power-on so confirm it or change it to 
// this default state. Start must be clearly defined
	DISP_E_0;
	Delay2ms;
// Set 8 bit communication
	DISP_E_0;
	DISP_INST;
	DISP_DB7_0;
	DISP_DB6_0;
	DISP_DB5_1;
	DISP_DB4_1;
	DISP_E_1;
	Delay1u;
	DISP_E_0;
	Delay50u;
	DISP_DB7_0;	// This is only second half or empty command
	DISP_DB6_0;
	DISP_DB5_0;
	DISP_DB4_0;
	DISP_E_1;
	Delay1u;
	DISP_E_0;
	Delay2ms;
// Set 4 bit communication
	DISP_E_0;
	DISP_DB7_0;
	DISP_DB6_0;
	DISP_DB5_1;
	DISP_DB4_0;
	DISP_E_1;
	Delay1u;
	DISP_E_0;
	Delay2ms;
// Set 2 line, 4 bit communication, font 5x7
	WriteIns(0x28);
// Set 2 line, 4 bit communication, font 5x7
	WriteIns(0x28);
// Display ON, Cursor OFF, Blink on cursor OFF
	WriteIns(0x0C);
// Reset display, Set cursor 0,0 position
	WriteIns(0x01);
	Delay2ms;
// Set cursor increment
	WriteIns(0x06);
}

// Write null terminated string to display
void WriteText(unsigned char *character)
{
   while(*character) {
   		WriteChar(*character);
		character++;
	}
}

// Set cursor position
void SetDispPos(uint8_t Xpos, uint8_t Ypos)
{
	if((Xpos>=MAXCHARS) || (Ypos>=MAXLINES)) return;
	if(Ypos) WriteIns(Xpos | 0xC0);
		else WriteIns(Xpos | 0x80);
}

// Clear display, set position 0,0
void ClrDisp()
{
// Reset display, Set cursor 0,0 position
	WriteIns(0x01);
	Delay2ms;
}

// Fill CGRAM with special characters - parameters:
// startchar - what character in CGRAM begins, numchar - how many chars, addrflash - pointer to flash table
void FillSpecChars(uint8_t startchar, uint8_t numchar, prog_uint8_t* addrflash)
{
uint8_t i,j;
  // Users' memory has 8 chars maximum 
  if((startchar+numchar)>8) return;
  // Start address in CGRAM
  WriteIns((startchar<<3) | 0x40);
  // Write every of "numchar" chars
  for(i=0; i<numchar; i++) 
	// Every char includes 8 lines
	for(j=0; j<8; j++) WriteChar(pgm_read_byte_near(addrflash++));
}

// Get tone counts from frequency table
unsigned long GetTone(int tonnumb)
{
  unsigned long RetValue;
  int auxTone,auxOctave;
  auxTone = tonnumb % 12;	// Only 12 tones in the table
  auxOctave = tonnumb / 12;	// What octave for this tone
  RetValue = pgm_read_dword_near(tones + auxTone);
  if(auxOctave) return(RetValue >> auxOctave);	// Every higher octave is /2
  else return(RetValue);
}

/* Measure 8 periods */
int Meas8per()
{
  unsigned char higcounter=0;

  TCCR1B = 0;
  TCNT1 = 0;
  if(TIFR1 & (1<<TOV1)) TIFR1 |= 0x01;
  // Clear Timer0
  TCNT0 = 0;
  // Clear OCR flags
  TIFR0 = 0x06;
  TCCR1B = 0x02;
  // Wait for measuring start (2 periods of input signal)
  do{
	if(TIFR1 & (1<<TOV1)) {
		TIFR1 |= 0x01;
		higcounter++;	// Every Timer1 overflow is counted
		if(higcounter > 20) return(0); // Timeout is 0.5 second - no input
	}
  }while (!(TIFR0 & (1<<OCF0A)));
  // Wait for measuring stop (8 periods of input signal - 2+8=10)
  TCNT1 = 0;
  if(TIFR1 & (1<<TOV1)) TIFR1 |= 0x01;
  higcounter = 0;
  do{
	if(TIFR1 & (1<<TOV1)) {
		TIFR1 |= 0x01;
		higcounter++;	// Every Timer1 overflow is counted
		if(higcounter > 10) return(0); // Timeout is 0.25 second - too long period
	}
  // Wait for 8 period
  } while (!(TIFR0 & (1<<OCF0B)));
  TCCR1B = 0; // Stop timer and restore measured time
  if(TIFR1 & (1<<TOV1)) higcounter++;	// Overflow during end condition

  measval = ((unsigned long int)higcounter << 16) | (unsigned long int)TCNT1;
  return(1);
}

/* Compares 4 measured values of 8 period, returns "1" when almost the same, 
   number of counts for 32 periods in sumval, number of tone in tonerange 
   and updown char for tuning arrow */
unsigned char Compare4vals()
{
  unsigned char i;
  long toler,delta,firstval; // allowed drift, diff. between measured values, ref. value

  // delta is about 3%
  if(Meas8per()) {firstval = sumval = measval; delta = measval >> 5;} else return(0);

  // Is it maximum +-delta from first value?
  for(i=0;i<3;i++) {
	if(Meas8per()) {sumval += measval; measval -= firstval; 
		if(measval>delta) return(0); if(measval<(-delta)) return(0);
		} else return(0);
  }
  // 32 periods * 2.5 MHz * 10 (1 decimal after point) = 800M
  freqres = (unsigned int) ((unsigned long)800000000 / sumval);

  // Find appropriate tone from table
  tonerange = 0;
  updown = ' ';
  // Lowest tone? Then it is finished.
  if(sumval > GetTone(0)) updown = 0; else {
	// Highest tone? Then it is finished.
	if(sumval < GetTone(59)) {updown = 1; tonerange=59;}
		else {
		// Find nearest higher tone
		for(i=0;i<59;i++) if(sumval < GetTone(i)) tonerange=i; else break;
		// Delta of 2 nearest tones / 2
		delta = GetTone(tonerange) - GetTone(tonerange+1);
		// 1/64 difference is O.K.
		toler = delta>>6;
		if(sumval > (GetTone(tonerange)-toler)) return(1);
		if(sumval < (GetTone(tonerange+1)+toler)) {tonerange++; return(1);}
		// Tune it up for first half of delta
		if(sumval < (GetTone(tonerange+1)+(delta>>1))) {tonerange++; updown = 0;}
		// or tune it down for second half of delta
		else updown = 1;
		}
	}
  return(1);
}

// ****************************
//    Main program function  //
// ****************************
int main(void) {
  int auxval;	// axiliary variable
  unsigned char nerrors=1;	// nerrors - counter unsuccessful meas, then :No input:
  unsigned char ton, nloop = 0xFF;	//ton-tone name char, nloop-Counter for Auto OFF

  /* Enable PORTB as output */
  DDRB = 0xFF;
  /* Enable PORTA as output except PA2,PA3*/
  DDRA = 0xF3;
  /* Enable pull-ups */
  PORTA = 0x0C;
  OFFTUNER;

  // Startup setting Timer 0
  // input from T0, rising edge
  TCCR0B = 0x07;
  // prepare measuring for 32 periods
  OCR0A = 2;
  OCR0B = 10;

  // Startup setting Timer 1 - no special setting

  // Watchdog to 4 sec
  WDTCSR |= (1<<WDCE) | (1<<WDE);
  WDTCSR = 0x28;

  // Display initialization and welcome message
  _delay_ms(10);
  InitDisplay();
  FillSpecChars(0,5,specchars);
  SetDispPos(1,0);
  WriteText(text);
  _delay_ms(100);
  ONTUNER;
  _delay_ms(2000);
  ClrDisp();
  // Constant chars for frequency displaying
  text[4] = '.';
  text[6] = 'H';
  text[7] = 'z';
  text[8] = 0;

  /***** Main loop of program *****/
  for(;;) {
	wdt_reset();
	if(Compare4vals()) {
		ClrDisp();
		// Write tone frequency on the bottom line
		SetDispPos(0,1);
		text[5] = (0x30 | freqres%10);	// Simple BCD conversion...
		freqres /= 10;
		text[3] = (0x30 | freqres%10);
		freqres /= 10;
		text[2] = (0x30 | freqres%10);
		freqres /= 10;	// ... with suppression of unsignificant zeroes
		if(freqres) text[1] = (0x30 | freqres%10); else text[1] = ' ';
		freqres /= 10;
		if(freqres) text[0] = (0x30 | (char)freqres); else text[0] = ' ';
		WriteText(text);
		// Write tone name on the upper line
		SetDispPos(2,0);
		// Index of tone name in the table
		auxval = (tonerange%12) << 1; // Tone name in table is 2 bytes
		// Upper case for first 24 tones only 
		ton = 0x0; if(tonerange>23) ton = 0x20;
		WriteChar(ton | names[auxval]);
		WriteChar(ton | names[auxval+1]);
		// Prepare octave index number from table
		ton = octind[tonerange/12];
		// And write last 2 tone symbols
		if(0x20 == names[auxval]) { WriteChar(ton);WriteChar(' '); }
		else { WriteChar('s');WriteChar(ton); }
		// And arrow for tuning
		WriteChar(updown);
		_delay_ms(1000);
		nerrors = 5;	// Valid output resets counter for error attempts
		nloop = 0xFF;	// Valid output resets AutoOFF
	} else { 
		// No reasonable input - signal it
		nerrors--;
		if(!nerrors) {
			ClrDisp();
			WriteText((unsigned char*)"No input");
		}
    }
    if(!(nloop--)) OFFTUNER;	// AutoOFF tuner in idle mode
  } /* loop forever */
  return 0; // Just for compiller
}

/* END */
 
