/* ========================================================================== */
/*   GTuner                                                                   */
/*   (c) 2015 SP                                                              */
/*                                                                            */
/*   Program for Guitar Tuner with PIC16F684 and 8x2 LCD                      */
/*   Clock frequency: Xtal 20 MHz                                             */
/* ========================================================================== */
// Version: 1.0 - first released (SourceBoost-BoostC 6.60)
#include <system.h>
#include <stdlib.h>

// Set configuration word 
#pragma OPTIMIZE "1p"
#pragma DATA 0x2007, _HS_OSC & _WDT_ON
#pragma CLOCK_FREQ 20000000

// Global variables (for quicker and smaller code)
long measval,sumval;
unsigned int freqres;	// Global variable with frequency
unsigned char updown, tonerange, text[9]={"Tuner\0"}; // Arrow for tuning, index nearest tone in table

// Tone names (2 chars only for each tone)
rom char *names = " CCi DDi E FFi GGi AAi H";

// Octave index
rom char *octind = {2,0x20,0x20,3,4};

// Graphic table for user characters in CGRAM in the LCD
rom char *specchars = {
// Up arrow
0b00000100,
0b00001110,
0b00011111,
0b00001110,
0b00001110,
0b00001110,
0b00001110,
0b00000000,
// Down arrow
0b00001110,
0b00001110,
0b00001110,
0b00001110,
0b00011111,
0b00001110,
0b00000100,
0b00000000,
// Lower 1
0b00000000,
0b00000000,
0b00000010,
0b00000110,
0b00000010,
0b00000010,
0b00000010,
0b00000000,
// Upper 1
0b00000010,
0b00000110,
0b00000010,
0b00000010,
0b00000010,
0b00000000,
0b00000000,
0b00000000,
// Upper 2
0b00000110,
0b00001001,
0b00000010,
0b00000100,
0b00001111,
0b00000000,
0b00000000,
0b00000000,
};

// 1 octave tone table (number of 2,5 MHz impulses in 32 periods)
rom char *tones = {0x25,0x53,0xA4,0x23,0x3B,0x55,0x21,0x41,0x1C,
0x1F,0x63,0x4C,0x1D,0xA0,0x53,0x1B,0xF6,0xA6,0x1A,0x64,0xDC,0x18,0xE9,0xA1,
0x17,0x83,0xAD,0x16,0x31,0xD1,0x14,0xF2,0xEB,0x13,0xC5,0xED};

//******** Bit manipulation definitions ***********
#define ONTUNER set_bit(portc,5)
#define OFFTUNER clear_bit(portc,5)
// LCD 2x8 chars, Write only, 4 bits bus
#define DISP_E_0 clear_bit(portc,4)  // LCD signals definition
#define DISP_E_1 set_bit(portc,4)
#define DISP_INST clear_bit(portc,3)
#define DISP_DATA set_bit(portc,3)
#define DISP_DB7_0 clear_bit(porta,0)
#define DISP_DB7_1 set_bit(porta,0)
#define DISP_DB6_0 clear_bit(porta,1)
#define DISP_DB6_1 set_bit(porta,1)
#define DISP_DB5_0 clear_bit(portc,1)
#define DISP_DB5_1 set_bit(portc,1)
#define DISP_DB4_0 clear_bit(portc,2)
#define DISP_DB4_1 set_bit(portc,2)
// Delay for display instructions
#define Delay1u delay_us(1)
#define Delay50u delay_us(50)
#define Delay2ms delay_ms(2)

typedef unsigned char uint8_t;

// Display type - define MAXCHARS x MAXLINES
#define MAXCHARS 8
#define MAXLINES 2
// ************** Display ******************/
// Write data to display - 4 bits bus
void WriteCom(char codedat)
{
	if(codedat & 0x80) DISP_DB7_1; else DISP_DB7_0;
	if(codedat & 0x40) DISP_DB6_1; else DISP_DB6_0;
	if(codedat & 0x20) DISP_DB5_1; else DISP_DB5_0;
	if(codedat & 0x10) DISP_DB4_1; else DISP_DB4_0;
	// Write upper nibble
	DISP_E_1;
	Delay1u;
	DISP_E_0;
	if(codedat & 0x08) DISP_DB7_1; else DISP_DB7_0;
	if(codedat & 0x04) DISP_DB6_1; else DISP_DB6_0;
	if(codedat & 0x02) DISP_DB5_1; else DISP_DB5_0;
	if(codedat & 0x01) DISP_DB4_1; else DISP_DB4_0;
	// Write lower nibble
	DISP_E_1;
	Delay1u;
	DISP_E_0;
	Delay50u;
}

// Write instruction code to display
void WriteIns(uint8_t codein)
{
	DISP_INST;	// Set display for instruction
	WriteCom(codein);
}

// Write data (char) to display
void WriteChar(uint8_t codein)
{
	DISP_DATA;	// Set display for data
	WriteCom(codein);
}

void InitDisplay()
{
// Display is 8bit after power-on so confirm it or change it to 
// this default state. Start must be clearly defined
	DISP_E_0;
	Delay2ms;
// Set 8 bit communication
	DISP_E_0;
	DISP_INST;
	DISP_DB7_0;
	DISP_DB6_0;
	DISP_DB5_1;
	DISP_DB4_1;
	DISP_E_1;
	Delay1u;
	DISP_E_0;
	Delay50u;
	DISP_DB7_0;	// This is only second half or empty command
	DISP_DB6_0;
	DISP_DB5_0;
	DISP_DB4_0;
	DISP_E_1;
	Delay1u;
	DISP_E_0;
	Delay2ms;
// Set 4 bit communication
	DISP_E_0;
	DISP_DB7_0;
	DISP_DB6_0;
	DISP_DB5_1;
	DISP_DB4_0;
	DISP_E_1;
	Delay1u;
	DISP_E_0;
	Delay2ms;
// Set 2 line, 4 bit communication, font 5x7
	WriteIns(0x28);
// Set 2 line, 4 bit communication, font 5x7
	WriteIns(0x28);
// Display ON, Cursor OFF, Blink on cursor OFF
	WriteIns(0x0C);
// Reset display, Set cursor 0,0 position
	WriteIns(0x01);
	Delay2ms;
// Set cursor increment
	WriteIns(0x06);
}

// Write null terminated string to display
void WriteText(char *character)
{
   while(*character) {
   		WriteChar(*character);
		character++;
	}
}

// Set cursor position
void SetDispPos(uint8_t Xpos, uint8_t Ypos)
{
	if((Xpos>=MAXCHARS) || (Ypos>=MAXLINES)) return;
	if(Ypos) WriteIns(Xpos | 0xC0);
		else WriteIns(Xpos | 0x80);
}

// Clear display, set position 0,0
void ClrDisp()
{
// Reset display, Set cursor 0,0 position
	WriteIns(0x01);
	Delay2ms;
}

// Fill CGRAM with special characters - parameters:
// startchar - what character in CGRAM begins, numchar - how many chars, addrflash - pointer to flash table
void FillSpecChars(uint8_t startchar, uint8_t numchar)
{
uint8_t i,j,addrflash = 0;
  // Users' memory has 8 chars maximum 
  if((startchar+numchar)>8) return;
  // Start address in CGRAM
  WriteIns((startchar<<3) | 0x40);
  // Write every of "numchar" chars
  for(i=0; i<numchar; i++) 
	// Every char includes 8 lines
	for(j=0; j<8; j++) WriteChar(specchars[addrflash++]);
}

// Get constant from memory table
unsigned long pgm_read_dword(unsigned char indtab)
{
  unsigned long output;
  unsigned char tablerow,tablecol;

  tablerow = indtab / 12;
  tablecol = (indtab % 12)*3;
  
  output = ((unsigned long)tones[tablecol])<<16 | ((unsigned long)tones[tablecol+1])<<8 |((unsigned long)tones[tablecol+2]);	
  
  if(tablerow) output = output >> (tablerow); // Div it for higher octave
  return(output);
}

/* Measure 8 periods */
unsigned char Meas8per()
{
//  unsigned long pomresult,delta;
  unsigned char higcounter=0;

  tmr1l = 0; tmr1h = 0; // clear TMR1 and its IF and start it
  clear_bit(pir1,0);
  higcounter = 0;
  set_bit(t1con,0);
  // Start Timer0
  tmr0 = 0xF6; //DE;
  clear_bit(intcon,2);	// clear TMR0 IF
  // wait for input signal
  do{
	if(test_bit(pir1,0)) {
		clear_bit(pir1,0);
		higcounter++;
		if(higcounter > 20) return(0);  // (16bit * 20) / 2.5 MHz = 0.5 sec
	}
  }while (!test_bit(tmr0,3)); // 2 input periods for stabilization - 0xF8 in TMR0
  // Start counting now
  tmr1l = 0; tmr1h = 0;
  clear_bit(pir1,0);
  higcounter = 0;
  do{
	if(test_bit(pir1,0)) {
		clear_bit(pir1,0);
		higcounter++;
		if(higcounter > 10) return(0); // Time-out 0.25 sec - minimum 32 Hz
	}
  // Wait for 8 periods
  } while (!(test_bit(intcon,2)));
  clear_bit(t1con,0);
  if(test_bit(pir1,0)) higcounter++;	// last overflow during end condition?
  measval = ((unsigned long)higcounter)<<16 | ((unsigned long)tmr1h)<<8 |((unsigned long)tmr1l);	
  return(1);
}

/* Compares 4 measured values of 8 period, returns "1" when almost the same, 
   number of counts for 32 periods in sumval, number of tone in tonerange 
   and updown char for tuning arrow */
unsigned char Compare4vals()
{
  unsigned char i;
  long delta,firstval; // allowed diff. between measured values, ref. value 

  // delta is about 3%
  if(Meas8per()) {firstval = sumval = measval; delta = measval >> 5;} else return(0);

  for(i=0;i<3;i++) {
	if(Meas8per()) {sumval += measval; measval -= firstval; 
		if(measval>delta) return(0); if(measval<(-delta)) return(0);
		} else return(0);
  }
  // 32 periods * 2.5 MHz * 10 (1 decimal after point) = 800M
  freqres = (unsigned int) ((unsigned long)800000000 / sumval);

  // Find appropriate tone from table
  tonerange = 0;
  updown = ' ';
  // Lowest tone? Then it is finished.
  if(sumval > pgm_read_dword(0)) updown = 0; else {
  	// Highest tone? Then it is finished.
  	if(sumval < pgm_read_dword(59)) {updown = 1; tonerange=59;}
		else {
		// Find nearest higher tone
		for(i=0;i<59;i++) if(sumval < pgm_read_dword(i)) tonerange=i; else break;
		// Delta of 2 nearest tones / 2
		delta = pgm_read_dword(tonerange) - pgm_read_dword(tonerange+1);
		// 1/64 difference is O.K.
		if(sumval > (pgm_read_dword(tonerange)-(delta>>6))) return(1);
		if(sumval < (pgm_read_dword(tonerange+1)+(delta>>6))) {tonerange++; return(1);}
		// Tune it up for first half of delta
		if(sumval < (pgm_read_dword(tonerange+1)+(delta>>1))) {tonerange++; updown = 0;}
		// or tune it down for second half of delta
		else updown = 1;
		}
	}
  return(1);
}

// *************************************************************** //
void main()
{
  unsigned char auxval, nerrors=1;	// nerrors - counter unsuccessful meas, then :No input:
  unsigned char ton, nloop = 0xFF;	// nloop - Counter for Auto OFF

  // Turn A/D and comparator off
  ansel = 0x00;
  cmcon0 = 0x07;
  // Set Port I/O
  trisa = 0b00000100;
  trisc = 0b00000001;

  // configure Timer0 - no special init

  // Init TIMER1 
  t1con = 0b00010001; // Timer control register

  // Display initialization and welcome message
  delay_ms(10);
  InitDisplay();
  FillSpecChars(0,5);
  SetDispPos(1,0);
  WriteText(text);
  delay_ms(100);
  ONTUNER;
  delay_s(2);
  ClrDisp();
  // Constant chars for frequency displaying
  text[4] = '.';
  text[6] = 'H';
  text[7] = 'z';
  text[8] = 0;

  /***** Main loop of program *****/
  for(;;) {
	clear_wdt();
    if(Compare4vals()) {
    	ClrDisp();
		// Write tone frequency on the bottom line
		SetDispPos(0,1);
		text[5] = (0x30 | freqres%10);	// Simple BCD conversion...
		freqres /= 10;
		text[3] = (0x30 | freqres%10);
		freqres /= 10;
		text[2] = (0x30 | freqres%10);
		freqres /= 10;	// ... with suppression of unsignificant zeroes
		if(freqres) text[1] = (0x30 | freqres%10); else text[1] = ' ';
		freqres /= 10;
		if(freqres) text[0] = (0x30 | (char)freqres); else text[0] = ' ';
		WriteText(text);
		// Write tone name on the upper line
		SetDispPos(2,0);
		// Index of tone name in the table
		auxval = (tonerange%12) << 1; // Tone name is 2 bytes
		// Upper case for first 24 tones only 
		if(tonerange>23) ton = 0x20; else ton = 0x0;
		WriteChar(ton | names[auxval]);
		WriteChar(ton | names[auxval+1]);
		// Prepare octave index number
		ton = octind[tonerange/12];
		// And write last 2 tone symbols
		if(0x20 == names[auxval]) { WriteChar(ton);WriteChar(' '); }
		else { WriteChar('s');WriteChar(ton); }
		// And arrow for tuning
		WriteChar(updown);
		delay_s(1);
		nerrors = 5;	// Valid output resets counter for error attempts
		nloop = 0xFF;	// Valid output resets AutoOFF
	} else { 
		// No reasonable input - signal it
		nerrors--;
		if(!nerrors) {
			ClrDisp();
			WriteText("No input");
		}
    }
    if(!(nloop--)) OFFTUNER;	// AutoOFF tuner in idle mode
  } /* loop forever */
}
